@@ -11,9 +11,11 @@
 
 use crate::{
     errors::RunIgnoredParseError,
+    list::RustTestArtifact,
     partition::{Partitioner, PartitionerBuilder},
 };
 use aho_corasick::AhoCorasick;
+use nextest_filtering::FilteringExpr;
 use nextest_metadata::{FilterMatch, MismatchReason};
 use std::{fmt, str::FromStr};
 
@@ -75,6 +77,7 @@ pub struct TestFilterBuilder {
     run_ignored: RunIgnored,
     partitioner_builder: Option<PartitionerBuilder>,
     name_match: NameMatch,
+    exprs: Vec<FilteringExpr>,
 }
 
 #[derive(Clone, Debug)]
@@ -91,16 +94,19 @@ impl TestFilterBuilder {
         run_ignored: RunIgnored,
         partitioner_builder: Option<PartitionerBuilder>,
         patterns: &[impl AsRef<[u8]>],
+        exprs: Vec<FilteringExpr>,
     ) -> Self {
         let name_match = if patterns.is_empty() {
             NameMatch::MatchAll
         } else {
             NameMatch::MatchSet(Box::new(AhoCorasick::new_auto_configured(patterns)))
         };
+
         Self {
             run_ignored,
             partitioner_builder,
             name_match,
+            exprs,
         }
     }
 
@@ -110,6 +116,7 @@ impl TestFilterBuilder {
             run_ignored,
             partitioner_builder: None,
             name_match: NameMatch::MatchAll,
+            exprs: Vec::new(),
         }
     }
 
@@ -137,46 +144,89 @@ pub struct TestFilter<'builder> {
 
 impl<'filter> TestFilter<'filter> {
     /// Returns an enum describing the match status of this filter.
-    pub fn filter_match(&mut self, test_name: &str, ignored: bool) -> FilterMatch {
+    pub fn filter_match(
+        &mut self,
+        test_binary: &RustTestArtifact<'_>,
+        test_name: &str,
+        ignored: bool,
+    ) -> FilterMatch {
+        self.filter_ignored_mismatch(ignored)
+            .or_else(|| self.filter_name_mismatch(test_name))
+            .or_else(|| self.filter_expression_mismatch(test_binary, test_name))
+            // Note that partition-based filtering MUST come after all other kinds of filtering,
+            // so that count-based bucketing applies after ignored, name and expression matching.
+            // This also means that mutable count state must be maintained by the partitioner.
+            .or_else(|| self.filter_partition_mismatch(test_name))
+            .unwrap_or(FilterMatch::Matches)
+    }
+
+    fn filter_ignored_mismatch(&self, ignored: bool) -> Option<FilterMatch> {
         match self.builder.run_ignored {
             RunIgnored::IgnoredOnly => {
                 if !ignored {
-                    return FilterMatch::Mismatch {
+                    return Some(FilterMatch::Mismatch {
                         reason: MismatchReason::Ignored,
-                    };
+                    });
                 }
             }
             RunIgnored::Default => {
                 if ignored {
-                    return FilterMatch::Mismatch {
+                    return Some(FilterMatch::Mismatch {
                         reason: MismatchReason::Ignored,
-                    };
+                    });
                 }
             }
             _ => {}
-        };
+        }
+        None
+    }
 
+    fn filter_name_mismatch(&self, test_name: &str) -> Option<FilterMatch> {
         let string_match = match &self.builder.name_match {
             NameMatch::MatchAll => true,
             NameMatch::MatchSet(set) => set.is_match(test_name),
         };
-        if !string_match {
-            return FilterMatch::Mismatch {
+        if string_match {
+            None
+        } else {
+            Some(FilterMatch::Mismatch {
                 reason: MismatchReason::String,
-            };
+            })
         }
+    }
 
+    fn filter_expression_mismatch(
+        &self,
+        test_binary: &RustTestArtifact<'_>,
+        test_name: &str,
+    ) -> Option<FilterMatch> {
+        let accepted = self.builder.exprs.is_empty()
+            || self
+                .builder
+                .exprs
+                .iter()
+                .any(|expr| expr.includes(test_binary.package.id(), test_name));
+
+        match accepted {
+            false => Some(FilterMatch::Mismatch {
+                reason: MismatchReason::Expression,
+            }),
+            true => None,
+        }
+    }
+
+    fn filter_partition_mismatch(&mut self, test_name: &str) -> Option<FilterMatch> {
         let partition_match = match &mut self.partitioner {
             Some(partitioner) => partitioner.test_matches(test_name),
             None => true,
         };
-        if !partition_match {
-            return FilterMatch::Mismatch {
+        if partition_match {
+            None
+        } else {
+            Some(FilterMatch::Mismatch {
                 reason: MismatchReason::Partition,
-            };
+            })
         }
-
-        FilterMatch::Matches
     }
 }
 
@@ -189,20 +239,20 @@ mod tests {
         #[test]
         fn proptest_empty(test_names in vec(any::<String>(), 0..16)) {
             let patterns: &[String] = &[];
-            let test_filter = TestFilterBuilder::new(RunIgnored::Default, None, patterns);
-            let mut single_filter = test_filter.build();
+            let test_filter = TestFilterBuilder::new(RunIgnored::Default, None, patterns, Vec::new());
+            let single_filter = test_filter.build();
             for test_name in test_names {
-                prop_assert!(single_filter.filter_match(&test_name, false).is_match());
+                prop_assert!(single_filter.filter_name_mismatch(&test_name).is_none());
             }
         }
 
         // Test that exact names match.
         #[test]
         fn proptest_exact(test_names in vec(any::<String>(), 0..16)) {
-            let test_filter = TestFilterBuilder::new(RunIgnored::Default, None, &test_names);
-            let mut single_filter = test_filter.build();
+            let test_filter = TestFilterBuilder::new(RunIgnored::Default, None, &test_names, Vec::new());
+            let single_filter = test_filter.build();
             for test_name in test_names {
-                prop_assert!(single_filter.filter_match(&test_name, false).is_match());
+                prop_assert!(single_filter.filter_name_mismatch(&test_name).is_none());
             }
         }
 
@@ -218,10 +268,10 @@ mod tests {
                 patterns.push(substring);
             }
 
-            let test_filter = TestFilterBuilder::new(RunIgnored::Default, None, &patterns);
-            let mut single_filter = test_filter.build();
+            let test_filter = TestFilterBuilder::new(RunIgnored::Default, None, &patterns, Vec::new());
+            let single_filter = test_filter.build();
             for test_name in test_names {
-                prop_assert!(single_filter.filter_match(&test_name, false).is_match());
+                prop_assert!(single_filter.filter_name_mismatch(&test_name).is_none());
             }
         }
 
@@ -234,9 +284,9 @@ mod tests {
         ) {
             prop_assume!(!substring.is_empty() && !(prefix.is_empty() && suffix.is_empty()));
             let pattern = prefix + &substring + &suffix;
-            let test_filter = TestFilterBuilder::new(RunIgnored::Default, None, &[&pattern]);
-            let mut single_filter = test_filter.build();
-            prop_assert!(!single_filter.filter_match(&substring, false).is_match());
+            let test_filter = TestFilterBuilder::new(RunIgnored::Default, None, &[&pattern], Vec::new());
+            let single_filter = test_filter.build();
+            prop_assert!(single_filter.filter_name_mismatch(&substring).is_some());
         }
     }
 
